import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.internal.model.builtin.builtinsymbol import IBuiltinSymbol
from boa3.internal.model.callable import Callable
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IBuiltinCallable(Callable, IBuiltinSymbol, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(args, vararg, kwargs, defaults, return_type)
        self._identifier = identifier
        self._generated_opcode = None
        self.defined_by_entry = False  # every builtin symbol must have this variable set as False

    def generate_opcodes(self, code_generator):
        """
        Generate the Neo VM opcodes for the method.

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        self.generate_internal_opcodes(code_generator)

    def generate_internal_opcodes(self, code_generator):
        """
        Generate the Neo VM opcodes for the method.

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        pass

    def reset(self):
        # reset the opcodes to ensure the correct output when calling consecutive compilations
        self._generated_opcode = None
        self.reset_calls()

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def identifier(self) -> str:
        return self._identifier



from __future__ import annotations

from typing import List

from boa3.internal import constants
from boa3.internal.neo import from_hex_str
from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.network.payloads.signer import Signer as TestSigner
from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope


class Signer(TestSigner):
    def __init__(self, account: UInt160, scopes: WitnessScope = WitnessScope.CalledByEntry):
        super().__init__(account, scopes)

    def add_scope(self, scope: WitnessScope):
        if scope == WitnessScope.WitnessRules:
            raise NotImplementedError
        if scope > self._scopes:
            self._scopes = scope

    def add_permitted_contracts(self, contract_hashes: list):
        for contract_hash in contract_hashes:
            if isinstance(contract_hash, bytes):
                script_hash = UInt160(contract_hash)
            elif isinstance(contract_hash, str):
                script_hash = UInt160.from_string(contract_hash)
            else:
                script_hash = contract_hash

            if not isinstance(script_hash, UInt160):
                raise ValueError(f"Incorrect data type for allowed contract: {type(script_hash).__name__}")
            if script_hash not in self._allowed_contracts:
                self._allowed_contracts.append(script_hash)

    def add_permitted_groups(self, groups_pubkeys: List[bytes]):
        for group_pubkey in groups_pubkeys:
            if isinstance(group_pubkey, str):
                group = from_hex_str(group_pubkey)
            else:
                group = group_pubkey

            if not isinstance(group, bytes):
                raise ValueError(f"Incorrect data type for allowed group: {type(group).__name__}")
            elif len(group) != constants.SIZE_OF_ECPOINT:
                raise ValueError(f"Incorrect data for allowed group")
            else:
                group = bytes(group)

            if group not in self._allowed_groups:
                self._allowed_groups.append(group)

    def add_permitted_rules(self, custom_rules: list):
        raise NotImplementedError

    def set_permissions(self, permissions: list):
        if self._scopes == WitnessScope.CustomContracts:
            self.add_permitted_contracts(permissions)
        elif self._scopes == WitnessScope.CustomGroups:
            self.add_permitted_groups(permissions)
        elif self._scopes == WitnessScope.WitnessRules:
            self.add_permitted_rules(permissions)



from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from boa3.internal.model.builtin.builtincallable import IBuiltinCallable
from boa3.internal.model.builtin.classmethod import *
from boa3.internal.model.builtin.compile_time import *
from boa3.internal.model.builtin.contract import *
from boa3.internal.model.builtin.decorator import *
from boa3.internal.model.builtin.internal import *
from boa3.internal.model.builtin.interop.interop import Interop
from boa3.internal.model.builtin.math import *
from boa3.internal.model.builtin.method import *
from boa3.internal.model.callable import Callable
from boa3.internal.model.event import Event as EventSymbol
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.imports.package import Package
from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.math import Math
from boa3.internal.model.type.neo import *


class BoaPackage(str, Enum):
    CompileTime = 'compile_time'
    Contract = 'contract'
    Interop = 'interop'
    Type = 'type'
    TypeHelper = 'helper'
    VM = 'vm'


class Builtin:
    @classmethod
    def get_symbol(cls, symbol_id: str) -> Optional[Callable]:
        for method in cls._python_builtins:
            if isinstance(method, IBuiltinCallable) and method.identifier == symbol_id:
                return method

    @classmethod
    def get_by_self(cls, symbol_id: str, self_type: IType) -> Optional[Callable]:
        for name, method in vars(cls).items():
            if (isinstance(method, IBuiltinMethod)
                    and method.identifier == symbol_id
                    and method.validate_self(self_type)):
                return method

    # builtin method
    Abs = AbsMethod()
    Exit = ExitMethod()
    IsInstance = IsInstanceMethod()
    Len = LenMethod()
    NewEvent = CreateEventMethod()
    Max = MaxIntMethod()
    Min = MinIntMethod()
    Print = PrintMethod()
    ScriptHashMethod_ = ScriptHashMethod()
    StrSplit = StrSplitMethod()
    Sum = SumMethod()
    ToHexStr = ToHexStrMethod()

    # python builtin class constructor
    Bool = BoolMethod()
    ByteArray = ByteArrayMethod()
    ByteArrayEncoding = ByteArrayEncodingMethod()
    Exception = ExceptionMethod()
    IntByteString = IntByteStringMethod()
    IntInt = IntIntMethod()
    ListBytesString = ListBytesStringMethod()
    ListGeneric = ListGenericMethod()
    ListMapping = ListMappingMethod()
    ListSequence = ListSequenceMethod()
    Range = RangeMethod()
    Reversed = ReversedMethod()
    StrBool = StrBoolMethod()
    StrBytes = StrBytesMethod()
    StrInt = StrIntMethod()
    Super = SuperMethod()

    # python class method
    BytesStringIndex = IndexBytesStringMethod()
    BytesStringIsDigit = IsDigitMethod()
    BytesStringJoin = JoinMethod()
    BytesStringLower = LowerMethod()
    BytesStringStartswith = StartsWithMethod()
    BytesStringStrip = StripMethod()
    BytesStringUpper = UpperMethod()
    CountSequenceGeneric = CountSequenceGenericMethod()
    CountSequencePrimitive = CountSequencePrimitiveMethod()
    CountStr = CountStrMethod()
    Copy = CopyListMethod()
    SequenceAppend = AppendMethod()
    SequenceClear = ClearMethod()
    SequenceExtend = ExtendMethod()
    SequenceIndex = IndexSequenceMethod()
    SequenceInsert = InsertMethod()
    SequencePop = PopSequenceMethod()
    SequenceRemove = RemoveMethod()
    SequenceReverse = ReverseMethod()
    DictKeys = MapKeysMethod()
    DictPop = PopDictMethod()
    DictPopDefault = PopDictDefaultMethod()
    DictValues = MapValuesMethod()

    # custom class methods
    ConvertToBytes = ToBytesMethod
    ConvertToInt = ToIntMethod
    ConvertToStr = ToStrMethod
    ConvertToBool = ToBoolMethod

    # builtin decorator
    ClassMethodDecorator = ClassMethodDecorator()
    InstanceMethodDecorator = InstanceMethodDecorator()
    PropertyDecorator = PropertyDecorator()
    StaticMethodDecorator = StaticMethodDecorator()

    _python_builtins: List[IdentifiedSymbol] = [Abs,
                                                ByteArray,
                                                ByteArrayEncoding,
                                                BytesStringIndex,
                                                BytesStringIsDigit,
                                                BytesStringJoin,
                                                BytesStringLower,
                                                BytesStringStartswith,
                                                BytesStringStrip,
                                                BytesStringUpper,
                                                ClassMethodDecorator,
                                                ConvertToBool,
                                                ConvertToBytes,
                                                ConvertToInt,
                                                ConvertToStr,
                                                Copy,
                                                CountSequenceGeneric,
                                                CountSequencePrimitive,
                                                CountStr,
                                                DictKeys,
                                                DictValues,
                                                Exception,
                                                Exit,
                                                IsInstance,
                                                Len,
                                                Max,
                                                Min,
                                                Print,
                                                PropertyDecorator,
                                                Range,
                                                Reversed,
                                                ScriptHashMethod_,
                                                SequenceAppend,
                                                SequenceClear,
                                                SequenceExtend,
                                                SequenceIndex,
                                                SequenceInsert,
                                                SequencePop,
                                                SequenceRemove,
                                                SequenceReverse,
                                                StaticMethodDecorator,
                                                StrSplit,
                                                Sum,
                                                Super,
                                                ]

    @classmethod
    def interop_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        return {symbol.raw_identifier if hasattr(symbol, 'raw_identifier') else symbol.identifier: symbol
                for symbol in Interop.interop_symbols(package)}

    # boa builtin decorator
    ContractInterface = ContractDecorator()
    ContractMethodDisplayName = DisplayNameDecorator()
    Metadata = MetadataDecorator()
    Public = PublicDecorator()

    # boa builtin type
    Event = EventType
    UInt160 = UInt160Type.build()
    UInt256 = UInt256Type.build()
    ECPoint = ECPointType.build()
    NeoAccountState = NeoAccountStateType.build()
    Opcode = OpcodeType.build()
    Address = AddressType.build()
    BlockHash = BlockHashType.build()
    PublicKey = PublicKeyType.build()
    ScriptHashType_ = ScriptHashType.build()
    ScriptHashLittleEndian = ScriptHashLittleEndianType.build()
    TransactionId = TransactionIdType.build()

    # boa events
    Nep11Transfer = Nep11TransferEvent()
    Nep17Transfer = Nep17TransferEvent()

    # boa smart contract methods
    Abort = AbortMethod()
    Env = EnvProperty.build()

    # region boa builtin modules

    BuiltinMathCeil = DecimalCeilingMethod()
    BuiltinMathFloor = DecimalFloorMethod()

    MathModule = Package(identifier='math',
                         methods=[Math.Sqrt,
                                  BuiltinMathCeil,
                                  BuiltinMathFloor])

    _symbols = [Env]
    _modules = [MathModule]

    # endregion

    boa_builtins: List[IdentifiedSymbol] = []
    boa_builtins.extend(_modules)
    boa_builtins.extend(_symbols)

    metadata_fields: Dict[str, Union[type, Tuple[type]]] = {
        'name': str,
        'source': (str, type(None)),
        'supported_standards': list,
        'trusts': list,
        'author': (str, type(None)),
        'email': (str, type(None)),
        'description': (str, type(None)),
        'extras': dict
    }

    @classmethod
    def boa_symbols(cls) -> Dict[str, IdentifiedSymbol]:
        return {symbol.identifier: symbol for symbol in cls.boa_builtins}

    @classmethod
    def package_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        if package in BoaPackage.__members__.values():
            return {symbol.identifier: symbol for symbol in cls._boa_symbols[package]}

        return cls.boa_symbols()

    @classmethod
    def builtin_events(cls) -> List[EventSymbol]:
        lst: List[EventSymbol] = [event for event in cls.boa_builtins if isinstance(event, EventSymbol)]

        for symbols in cls._boa_symbols.values():
            lst.extend([event for event in symbols if isinstance(event, EventSymbol)])

        return lst

    _builtin_type_package_symbols = [ECPoint,
                                     UInt160,
                                     UInt256,
                                     Event,
                                     Address,
                                     BlockHash,
                                     PublicKey,
                                     ScriptHashType_,
                                     ScriptHashLittleEndian,
                                     TransactionId,
                                     Package(identifier=BoaPackage.TypeHelper,
                                             methods=[ConvertToBool,
                                                      ConvertToBytes,
                                                      ConvertToInt,
                                                      ConvertToStr,
                                                      ]
                                             )
                                     ]

    _boa_symbols: Dict[BoaPackage, List[IdentifiedSymbol]] = {
        BoaPackage.Contract: [Abort,
                              NeoAccountState,
                              Nep11Transfer,
                              Nep17Transfer,
                              ScriptHashMethod_,
                              ToHexStr,
                              ],
        BoaPackage.Interop: Interop.package_symbols,
        BoaPackage.Type: _builtin_type_package_symbols,
        BoaPackage.VM: [Opcode
                        ],
        BoaPackage.CompileTime: [ContractInterface,
                                 ContractMethodDisplayName,
                                 Metadata,
                                 NeoMetadataType,
                                 Public,
                                 NewEvent
                                 ]
    }

    _internal_methods = [InnerDeployMethod.instance()
                         ]
    internal_methods = {method.raw_identifier: method for method in _internal_methods}



from typing import Dict

from boa3.internal.model.builtin.method.strmethod import StrMethod
from boa3.internal.model.variable import Variable


class StrIntMethod(StrMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        args: Dict[str, Variable] = {
            'object': Variable(Type.int),
        }

        super().__init__(args)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.stdlib import ItoaMethod

        code_generator.convert_builtin_method_call(ItoaMethod(internal_call_args=1), is_internal=True)



from boa3.internal.model.builtin.method.printmethod import PrintMethod

from boa3.internal.model.type.itype import IType


class PrintByteStringMethod(PrintMethod):

    def __init__(self, arg_value: IType):
        from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType

        if not isinstance(arg_value, IType) or not isinstance(arg_value, IByteStringType):
            from boa3.internal.model.type.type import Type
            arg_value = Type.str

        super().__init__(arg_value)



from __future__ import annotations

__all__ = [
    'Iterator',
]


from typing import Any


class Iterator:
    """
    The iterator for smart contracts.
    """

    @property
    def value(self) -> Any:
        """
        Gets the element in the collection at the current position of the iterator.

        :return: the element in the collection at the current position of the iterator
        :rtype: Any
        """
        return None

    def next(self) -> bool:
        """
        Advances the iterator to the next element of the collection.

        >>> from boa3.builtin.interop import storage
        ... iterator = storage.find(b'prefix')
        ... iterator.next()
        True

        :return: true if it advanced, false if there isn't a next element
        :rtype: bool
        """
        pass

    def __next__(self):
        if self.next():
            return self.value
        raise StopIteration

    def __iter__(self):
        return self



__all__ = [
    'to_bool',
    'to_bytes',
    'to_int',
    'to_str',
]

from typing import Union


def to_bytes(value: Union[str, int]) -> bytes:
    """
    Converts a str or integer value to an array of bytes

    >>> to_bytes(65)
    b'A'

    >>> to_bytes('A')
    b'A'
    """
    pass


def to_str(value: bytes) -> str:
    """
    Converts a bytes value to a string.

    >>> to_str(b'A')
    'A'
    """
    pass


def to_int(value: bytes) -> int:
    """
    Converts a bytes value to the integer it represents.

    >>> to_int(b'A')
    65
    """
    pass


def to_bool(value: bytes) -> bool:
    """
    Return a bytes value to the boolean it represents.

    >>> to_bool(b'\\x00')
    False

    >>> to_bool(b'\\x01')
    True

    >>> to_bool(b'\\x02')
    True
    """
    pass



__all__ = [
    'get_account_by_name',
    'get_account_by_address',
    'get_account_by_identifier',
    'get_default_account',
    'get_genesis_block',
    'get_account_from_script_hash_or_id',
    'reset_neo_express_instance',
    'get_deployed_contracts',
    'get_transaction',
    'get_transaction_log',
    'get_latest_block',
    'get_block',
    'run_batch',
    'oracle_response'
]


import logging
from typing import List, Union

from filelock import FileLock

from boa3.internal import env
from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.neoxp import utils as neoxp_utils
from boa3_test.test_drive.neoxp.executor import NeoExpressWrapper
from boa3_test.test_drive.neoxp.model.neoxpconfig import NeoExpressConfig
from boa3_test.test_drive.neoxp.utils import (Account,
                                              Block,
                                              Contract,
                                              Transaction,
                                              TransactionLog,
                                              )

_NEOXP_CONFIG = NeoExpressConfig(f'{env.NEO_EXPRESS_INSTANCE_DIRECTORY}/default.neo-express')
_NEOXP_FILE_LOCK = FileLock(f'{env.TEST_RUNNER_DIRECTORY}/test-runner.lock')
logging.getLogger("filelock").setLevel(logging.INFO)


def get_account_by_name(account_name) -> Account:
    return neoxp_utils.get_account_by_name(_NEOXP_CONFIG, account_name)


def get_account_by_address(account_address: str) -> Account:
    return neoxp_utils.get_account_by_address(_NEOXP_CONFIG, account_address)


def get_account_by_identifier(account_identifier: str) -> Account:
    return neoxp_utils.get_account_by_identifier(_NEOXP_CONFIG, account_identifier)


def get_account_from_script_hash_or_id(script_hash_or_address: Union[bytes, str]) -> Account:
    return neoxp_utils.get_account_from_script_hash_or_id(_NEOXP_CONFIG, script_hash_or_address)


def get_default_account() -> Account:
    return _NEOXP_CONFIG.default_account


def get_address_version() -> int:
    return _NEOXP_CONFIG.version


def get_magic() -> int:
    return _NEOXP_CONFIG.magic


def get_genesis_block() -> Block:
    return _NEOXP_CONFIG.genesis_block


class _BoaNeoExpressWrapper(NeoExpressWrapper):
    @classmethod
    def _run_cli(cls, *neoxp_args: str):
        with _NEOXP_FILE_LOCK:
            cli_result = super()._run_cli(*neoxp_args)
        return cli_result


def reset_neo_express_instance(neoxp_path: str, check_point_file: str = None) -> str:
    return _BoaNeoExpressWrapper.reset_neo_express_instance(neoxp_path, check_point_file)


def get_deployed_contracts(neoxp_path: str, check_point_file: str = None) -> List[Contract]:
    return _BoaNeoExpressWrapper.get_deployed_contracts(neoxp_path, check_point_file)


def get_transaction(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> Transaction:
    return _BoaNeoExpressWrapper.get_transaction(neoxp_path, tx_hash, check_point_file)


def get_transaction_log(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None, contract_collection=None) -> TransactionLog:
    return _BoaNeoExpressWrapper.get_transaction_log(neoxp_path, tx_hash, check_point_file, contract_collection)


def get_latest_block(neoxp_path: str, check_point_file: str = None) -> Block:
    return _BoaNeoExpressWrapper.get_latest_block(neoxp_path, check_point_file)


def get_block(neoxp_path: str, block_hash_or_index: Union[UInt256, int] = None,
              check_point_file: str = None) -> Block:
    return _BoaNeoExpressWrapper.get_block(neoxp_path, block_hash_or_index, check_point_file)


def run_batch(neoxp_path: str, batch_path: str, reset: bool = False, check_point_file: str = None) -> str:
    return _BoaNeoExpressWrapper.run_batch(neoxp_path, batch_path, reset, check_point_file)


def oracle_response(neoxp_path: str, url: str, response_path: str, request_id: int = None,
                    check_point_file: str = None) -> List[UInt256]:
    return _BoaNeoExpressWrapper.oracle_response(neoxp_path, url, response_path, request_id, check_point_file)



from typing import Any, Dict, Optional, Sized

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class LenMethod(IBuiltinMethod):

    def __init__(self, collection_type: IType = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(collection_type, ICollectionType):
            collection_type = Type.sequence

        identifier = 'len'
        args: Dict[str, Variable] = {'__o': Variable(collection_type)}
        super().__init__(identifier, args, return_type=Type.int)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.SIZE)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['__o'].type):
            return self

        if isinstance(value, Sized) and len(value) == 1:
            value = value[0]
        if isinstance(value, ICollectionType):
            return LenMethod(value)
        return super().build(value)



from __future__ import annotations

from typing import Any, Dict

from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.primitive.inttype import IntType
from boa3.internal.neo3.network.payloads.verification import WitnessScope


class WitnessScopeType(IntType):
    """
    A class used to represent Neo interop WitnessScope type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'WitnessScope'

    @property
    def default_value(self) -> Any:
        return WitnessScope.NONE

    @classmethod
    def build(cls, value: Any = None) -> WitnessScopeType:
        if value is None or cls._is_type_of(value):
            return _WitnessScope

    @classmethod
    def _is_type_of(cls, value: Any = None):
        return isinstance(value, (WitnessScope, WitnessScopeType))

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the module with its name
        """
        from boa3.internal.model.variable import Variable

        _symbols = super().symbols
        _symbols.update({name: Variable(self) for name in WitnessScope.__members__.keys()})

        return _symbols

    def get_value(self, symbol_id) -> Any:
        """
        Gets the literal value of a symbol

        :return: the value if this type has this symbol. None otherwise.
        """
        if symbol_id in self.symbols:
            return WitnessScope.__members__[symbol_id]

        return None


_WitnessScope = WitnessScopeType()



from typing import List, Optional, Tuple

from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class CollectionMembership(BinaryOperation):
    """
    A class used to represent a collection membership operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.collection]

    def __init__(self, left: IType = Type.any, right: IType = None):
        self.operator: Operator = Operator.In

        if not isinstance(right, ICollectionType):
            right = Type.collection

        reference_type = right.key_type if isinstance(right, MappingType) else right.item_type
        if not reference_type.is_type_of(left) or (reference_type is Type.any and left is not Type.any):
            if Type.sequence.is_type_of(left):
                right = Type.collection.build(left)
            else:
                right = Type.collection.build(Type.list.build([left]))

        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if not isinstance(right, ICollectionType):
            return False

        if isinstance(right, PrimitiveType) and (right.is_type_of(left) or left.is_type_of(right)):
            return True

        if isinstance(right, MappingType):
            reference_type = right.key_type
        else:
            reference_type = right.item_type

        return reference_type.is_type_of(left)

    def _get_result(self, left: IType, right: IType) -> IType:
        return Type.bool

    def get_valid_operand_for_validation(self, left_operand: IType,
                                         right_operand: IType = None) -> Tuple[Optional[IType], Optional[IType]]:
        if isinstance(right_operand, ICollectionType):
            left = right_operand.item_type if not isinstance(right_operand, MappingType) else right_operand.key_type
            return left, right_operand

        return left_operand, self.right_type

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(Type.dict.stack_item)
        # if isinstance(arg1, dict)
        is_dict = code_generator.convert_begin_if()

        #   return value.has_key(arg0)
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.HASKEY, pop_from_stack=True, add_to_stack=[Type.bool])

        is_dict = code_generator.convert_begin_else(is_dict, is_internal=True)
        code_generator.duplicate_stack_top_item()

        # is_bytestr = isinstance(arg1, (str, bytes))
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(Type.str.stack_item)

        # limit = len(arg1)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.duplicate_stack_item(2)
        # if is_bytestr:
        is_bytestr = code_generator.convert_begin_if()

        #   limit = limit - len(arg0) + 1
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)

        code_generator.convert_end_if(is_bytestr)

        # index = 0
        code_generator.convert_literal(0)

        # while index < limit:
        while_start = code_generator.convert_begin_while()

        code_generator.duplicate_stack_item(5)
        code_generator.duplicate_stack_item(5)
        code_generator.duplicate_stack_item(3)
        #   aux = arg0

        code_generator.duplicate_stack_item(6)
        #   if is_bytestr:
        is_bytestr_while = code_generator.convert_begin_if()

        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        #       if arg1[index:len(arg0)] == arg0:
        break_if = code_generator.convert_begin_if()
        #           break
        code_generator.convert_loop_break()

        code_generator.convert_end_if(break_if, is_internal=True)
        is_bytestr_while = code_generator.convert_begin_else(is_bytestr_while, is_internal=True)

        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.convert_operation(BinaryOp.Eq, is_internal=True)
        #   elif arg1[index] == arg0:
        break_elif = code_generator.convert_begin_if()
        #       break
        code_generator.convert_loop_break()

        code_generator.convert_end_if(break_elif, is_internal=True)
        code_generator.convert_end_if(is_bytestr_while, is_internal=True)

        #   index += 1
        code_generator.insert_opcode(Opcode.INC)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # return limit > index
        #   if the value is found, index won't reach the limit
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        # remove auxiliary values from stack
        code_generator.swap_reverse_stack_items(4)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

        code_generator.convert_end_if(is_dict, is_internal=True)



from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ExtendMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(sequence_type, MutableSequenceType):
            sequence_type = Type.mutableSequence

        self_arg = Variable(sequence_type)
        item_arg = Variable(Type.sequence.build_collection(sequence_type.value_type))

        identifier = 'extend'
        args: Dict[str, Variable] = {'self': self_arg, 'item': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _arg_item(self) -> Variable:
        return self.args['item']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type
        iterator_type: IType = params[1].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        if not isinstance(iterator_type, SequenceType):
            return False
        return sequence_type.value_type.is_type_of(iterator_type.value_type)

    @property
    def stores_on_slot(self) -> bool:
        return True

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        expected_values = self._arg_item.type

        code_generator.duplicate_stack_item(2)
        code_generator.insert_type_check(StackItemType.Buffer)  # append opcode only works for array
        # if self is bytes
        if_bytes = code_generator.convert_begin_if()
        #   result = self + arg     # when it's bytearray, concatenates the value
        code_generator.convert_operation(BinaryOp.Concat.build(Type.bytearray, Type.bytearray))

        # else:
        else_bytes = code_generator.convert_begin_else(if_bytes, is_internal=True)
        #   while len(item) > 0
        code_generator.insert_opcode(Opcode.UNPACK, add_to_stack=[expected_values, Type.int])

        begin_while = code_generator.convert_begin_while()
        #       push the array to the top of the stack
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.INC)
        code_generator.duplicate_stack_item(expected_stack_item=expected_values)

        #       get the first value that wasn't appended yet
        code_generator.move_stack_item_to_top(3)

        #       self.append(item)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)

        #       update stack control
        code_generator.insert_opcode(Opcode.DEC)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        #   while there are items to add

        code_generator.convert_end_while(begin_while, while_condition, is_internal=True)
        #   clear stack
        for _ in self.args:
            code_generator.remove_stack_top_item()

        code_generator.convert_end_if(else_bytes, is_internal=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return ExtendMethod(value)
        return super().build(value)



__all__ = [
    'get_config_data',
    'get_account_by_name',
    'get_account_by_address',
    'get_account_by_identifier',
    'get_default_account',
    'get_genesis_block',
    'get_account_from_script_hash_or_id',
    'create_neo_express_instance',
    'reset_neo_express_instance',
    'get_deployed_contracts',
    'get_transaction',
    'get_transaction_log',
    'get_latest_block',
    'get_block',
    'run_batch',
    'oracle_response',
    'Account',
    'Block',
    'Contract',
    'NeoExpressConfig',
    'Transaction',
    'TransactionLog'
]

from typing import List, Union

from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.wallet import utils as wallet_utils
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.executor import NeoExpressWrapper
from boa3_test.test_drive.neoxp.model.neoxpconfig import NeoExpressConfig
from boa3_test.test_drive.testrunner.blockchain.block import TestRunnerBlock as Block
from boa3_test.test_drive.testrunner.blockchain.contract import TestRunnerContract as Contract
from boa3_test.test_drive.testrunner.blockchain.transaction import TestRunnerTransaction as Transaction
from boa3_test.test_drive.testrunner.blockchain.transactionlog import TestRunnerTransactionLog as TransactionLog


def get_config_data(neoxp_path: str) -> NeoExpressConfig:
    return NeoExpressWrapper.get_config_data(neoxp_path)


def get_account_by_name(neoxp_config: NeoExpressConfig, account_name) -> Account:
    return next((account for account in neoxp_config.accounts if account.name == account_name),
                None)


def get_account_by_address(neoxp_config: NeoExpressConfig, account_address: str) -> Account:
    return next((account for account in neoxp_config.accounts if account.address == account_address),
                None)


def get_account_by_identifier(neoxp_config: NeoExpressConfig, account_identifier: str) -> Account:
    return next((account for account in neoxp_config.accounts if (account.address == account_identifier
                                                                  or account.name == account_identifier)),
                None)


def get_default_account(neoxp_config: NeoExpressConfig) -> Account:
    return neoxp_config.default_account


def get_genesis_block(neoxp_config: NeoExpressConfig) -> Block:
    return neoxp_config.genesis_block


def get_account_from_script_hash_or_id(neoxp_config: NeoExpressConfig, script_hash_or_address: Union[bytes, str]) -> Account:
    if isinstance(script_hash_or_address, bytes):
        script_hash = script_hash_or_address
        address = wallet_utils.address_from_script_hash(script_hash, neoxp_config.version)
        account = get_account_by_address(neoxp_config, address)
    elif isinstance(script_hash_or_address, str):
        account = get_account_by_identifier(neoxp_config, script_hash_or_address)
        script_hash = wallet_utils.address_to_script_hash(script_hash_or_address, neoxp_config.version)
    else:
        raise TypeError(f"Invalid data type {type(script_hash_or_address)}. Expecting str or bytes")

    if not isinstance(account, Account):
        from boa3.internal.neo3.core.types import UInt160
        from boa3_test.test_drive.neoxp.model.neoxpaccount import NeoExpressAccount
        account = NeoExpressAccount(UInt160(script_hash), neoxp_config.version)

    return account


def create_neo_express_instance(neoxp_path: str) -> str:
    return NeoExpressWrapper.create_neo_express_instance(neoxp_path)


def reset_neo_express_instance(neoxp_path: str, check_point_file: str = None) -> str:
    return NeoExpressWrapper.reset_neo_express_instance(neoxp_path, check_point_file)


def get_deployed_contracts(neoxp_path: str, check_point_file: str = None) -> List[Contract]:
    return NeoExpressWrapper.get_deployed_contracts(neoxp_path, check_point_file)


def get_transaction(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> Transaction:
    return NeoExpressWrapper.get_transaction(neoxp_path, tx_hash, check_point_file)


def get_transaction_log(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None, contract_collection=None) -> TransactionLog:
    return NeoExpressWrapper.get_transaction_log(neoxp_path, tx_hash, check_point_file, contract_collection)


def get_latest_block(neoxp_path: str, check_point_file: str = None) -> Block:
    return NeoExpressWrapper.get_latest_block(neoxp_path, check_point_file)


def get_block(neoxp_path: str, block_hash_or_index: Union[UInt256, int] = None,
              check_point_file: str = None) -> Block:
    return NeoExpressWrapper.get_block(neoxp_path, block_hash_or_index, check_point_file)


def run_batch(neoxp_path: str, batch_path: str, reset: bool = False, check_point_file: str = None) -> str:
    return NeoExpressWrapper.run_batch(neoxp_path, batch_path, reset, check_point_file)


def oracle_response(neoxp_path: str, url: str, response_path: str, request_id: int = None,
                    check_point_file: str = None) -> List[UInt256]:
    return NeoExpressWrapper.oracle_response(neoxp_path, url, response_path, request_id, check_point_file)



from typing import List, Optional

from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class Concat(BinaryOperation):
    """
    A class used to represent a string concatenation operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.str, Type.bytes]

    def __init__(self, left: IType = Type.str, right: IType = None):
        self.operator: Operator = Operator.Plus
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if left == right and any(valid_type.is_type_of(left) for valid_type in self._valid_types):
            return True

        left_valid_type = self._get_valid_type(left)
        if left_valid_type is None:
            return False

        right_valid_type = self._get_valid_type(right)

        return (right_valid_type is not None
                and (left_valid_type == right_valid_type
                     or left_valid_type.is_type_of(right_valid_type)
                     or right_valid_type.is_type_of(left_valid_type))
                )

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    def _get_valid_type(self, operator_type: IType) -> Optional[IType]:
        return next((valid_type for valid_type in self._valid_types
                     if valid_type.is_type_of(operator_type)), None)

    def generate_opcodes(self, code_generator):
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        super().generate_opcodes(code_generator)
        if self.result.stack_item != StackItemType.Buffer:
            code_generator.convert_cast(self.result, is_internal=True)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.CAT)



from typing import Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable


class StorageMapPutMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype import _StorageMap
        from boa3.internal.model.type.type import Type

        identifier = 'put'
        storage_value_type = Type.union.build([Type.bytes,
                                               Type.int,
                                               Type.str,
                                               ])

        args: Dict[str, Variable] = {'self': Variable(_StorageMap),
                                     'key': Variable(Type.bytes),
                                     'value': Variable(storage_value_type)}

        super().__init__(identifier, args, return_type=Type.none)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.operation.binaryop import BinaryOp

        # actual_key = self._prefix + key
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        code_generator.convert_get_item(index_inserted_internally=True)  # self._prefix

        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        # actual_context = self._context
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_literal(0)
        code_generator.convert_get_item(index_inserted_internally=True)  # self._context

        # return put(actual_key, actual_context)
        code_generator.convert_builtin_method_call(Interop.StoragePut, is_internal=True)



from typing import Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable


class StorageContextCreateMapMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.builtin.interop.storage.storagecontext.storagecontexttype import _StorageContext
        from boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype import _StorageMap
        from boa3.internal.model.type.type import Type

        identifier = 'create_map'
        byte_string_type = Type.bytes

        args: Dict[str, Variable] = {'self': Variable(_StorageContext),
                                     'prefix': Variable(byte_string_type)}

        super().__init__(identifier, args, return_type=_StorageMap)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype import _StorageMap as StorageMapType
        code_generator.convert_builtin_method_call(StorageMapType.constructor_method(), is_internal=True)



from typing import Dict

from boa3.internal.model.builtin.method.listmethod import ListMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ListBytesStringMethod(ListMethod):

    def __init__(self, value: IType = None):
        from boa3.internal.model.type.type import Type

        if value is None:
            value = Type.bytes

        args: Dict[str, Variable] = {
            'value': Variable(value),
        }

        return_type = Type.list.build_collection(value if value is Type.str else Type.int)

        super().__init__(args, return_type)

    def generate_pack_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type

        # index = len(value) - 1
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)

        # while index >= 0
        start_address = code_generator.convert_begin_while()

        #   value[index] to stack
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)

        if self._arg_value.type is Type.str:
            code_generator.convert_cast(Type.str, is_internal=True)

        #   reorganize stack
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.insert_opcode(Opcode.DEC)

        #   while condition
        condition_address = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.SIGN)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)

        code_generator.convert_end_while(start_address, condition_address, is_internal=True)

        # clear stack
        code_generator.remove_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)



from typing import Dict

from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.icontractcommand import IContractCommand


class ContractInvokeCommand(IContractCommand):
    def __init__(self, invoke_file: str, account: Account,
                 witness_scope: WitnessScope = None,
                 results_only: bool = False,
                 additional_gas: int = 0,
                 wallet_password: str = None,
                 trace: bool = False,
                 json_output: bool = False,
                 neo_express_data_file: str = None):

        self.invoke_file = invoke_file
        self.account = account

        self.witness_scope = witness_scope
        self.password = wallet_password

        self.results = True if not isinstance(account, Account) else results_only
        self.additional_gas = additional_gas if additional_gas > 0 else 0
        self.input = neo_express_data_file
        self.trace = trace
        self.json_output = json_output

        super().__init__('invoke', [invoke_file, account.get_identifier()])

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if isinstance(self.witness_scope, WitnessScope):
            options['--witness-scope'] = self.witness_scope.name
        if self.results:
            options['--results'] = ''
        if self.additional_gas > 0:
            options['--gas'] = utils.stringify_asset_quantity(self.additional_gas, utils.GAS_DECIMALS)
        if isinstance(self.password, str):
            options['--password'] = self.password
        if self.trace:
            options['--trace'] = ''
        if self.json_output:
            options['--json'] = ''

        return options



if __name__ == '__main__':
    import sys
    import os.path
    import shutil

    try:
        import boa3
    except ImportError as e:

        project_root = os.path.abspath(f'{os.path.dirname(__file__)}/../..')
        sys.path.append(project_root)
    finally:
        from boa3.internal import env
        from boa3_test.tests import boa_test
        from boa3_test.tests.test_suite import *
        from boa3_test.test_drive import utils

    neo_express_dir = env.NEO_EXPRESS_INSTANCE_DIRECTORY
    env_test_runner_dir = env.TEST_RUNNER_DIRECTORY

    # avoids to have multiple applications accessing the same config file
    test_runner_dir = os.path.abspath('./test-runner')
    neo_express_exec_dir = f'{test_runner_dir}/{utils.create_custom_id("neoxp", use_time=False)}'

    try:
        if not os.path.exists(test_runner_dir):
            os.mkdir(test_runner_dir)

        if not os.path.exists(neo_express_exec_dir):
            os.mkdir(neo_express_exec_dir)

        neo_express_exec_file = f'{neo_express_exec_dir}/default.neo-express'
        shutil.copy(f'{neo_express_dir}/default.neo-express', neo_express_exec_file)
        env.NEO_EXPRESS_INSTANCE_DIRECTORY = neo_express_exec_dir
        env.TEST_RUNNER_DIRECTORY = test_runner_dir
        boa_test.USE_UNIQUE_NAME = True

        suite = AsyncTestSuite()
        discover_path = f'{env.PROJECT_ROOT_DIRECTORY}/boa3_test/'
        test_discover = unittest.loader.defaultTestLoader.discover(discover_path,
                                                                   top_level_dir=env.PROJECT_ROOT_DIRECTORY,
                                                                   )

        for test in list_of_tests_gen(test_discover):
            suite.addTest(test)

        print(f'Found {suite.countTestCases()} tests\n')

        test_result = unittest.TextTestRunner(verbosity=2,
                                              resultclass=CustomTestResult,
                                              ).run(suite)

        sys.exit(not test_result.wasSuccessful())
    finally:
        # set environment variables back to their starting state
        env.NEO_EXPRESS_INSTANCE_DIRECTORY = neo_express_dir
        env.TEST_RUNNER_DIRECTORY = env_test_runner_dir
        boa_test.USE_UNIQUE_NAME = False
        # clear test directories
        shutil.rmtree(neo_express_exec_dir)
        if len(os.listdir(test_runner_dir)) == 0:
            os.rmdir(test_runner_dir)

    sys.exit(0)



__all__ = [
    'AddressType',
    'BlockHashType',
    'OpcodeType',
    'PublicKeyType',
    'ScriptHashLittleEndianType',
    'ScriptHashType',
    'TransactionIdType',
]

from boa3.internal.model.type.neo.addresstype import AddressType
from boa3.internal.model.type.neo.blockhashtype import BlockHashType
from boa3.internal.model.type.neo.opcodetype import OpcodeType
from boa3.internal.model.type.neo.publickeytype import PublicKeyType
from boa3.internal.model.type.neo.scripthashlittleendiantype import ScriptHashLittleEndianType
from boa3.internal.model.type.neo.scripthashtype import ScriptHashType
from boa3.internal.model.type.neo.transactionidtype import TransactionIdType



from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ECPointMethod(IBuiltinMethod):

    def __init__(self, return_type: ECPointType, argument_type: IType = None):
        from boa3.internal.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'ECPoint'
        args: Dict[str, Variable] = {'arg': Variable(argument_type)}

        super().__init__(identifier, args, return_type=return_type)

    @property
    def _arg_arg(self) -> Variable:
        return self.args['arg']

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_arg.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_arg.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.internal.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.internal.model.type.type import Type
        return Type.bytes.is_type_of(param_type)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal import constants
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.operation.unaryop import UnaryOp

        # if (arg is not None
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(None)
        code_generator.convert_operation(UnaryOp.Not, is_internal=True)

        is_arg_null = code_generator.convert_begin_if()
        #       or len(arg) != ECPOINT_SIZE):
        code_generator.convert_cast(self.return_type, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(constants.SIZE_OF_ECPOINT)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)

        else_address = code_generator.convert_begin_else(is_arg_null, True, is_internal=True)
        code_generator.change_jump(else_address, Opcode.JMPIF)
        #   raise exception
        code_generator.convert_raise_exception(is_internal=True)
        code_generator.convert_end_if(else_address, is_internal=True)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_arg.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ECPointMethod(self.return_type, value)
        return super().build(value)



import re
from typing import Dict, List

from boa3.internal.analyser import supportedstandard
from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.exception import CompilerError
from boa3.internal.model.event import Event
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol


class StandardAnalyser(IAstAnalyser):
    """
    This class is responsible for the checking if the given contract standards are implemented

    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: Dict[str, ISymbol],
                 log: bool = False, fail_fast: bool = True):
        from boa3.builtin.compile_time import NeoMetadata

        super().__init__(analyser.ast_tree, analyser.filename, analyser.root, log=log, fail_fast=fail_fast)

        self.symbols: Dict[str, ISymbol] = symbol_table

        if isinstance(analyser.metadata, NeoMetadata):
            # filter only strings
            analyser.metadata.supported_standards = [standard for standard in analyser.metadata.supported_standards
                                                     if isinstance(standard, str)]
            standards = analyser.metadata.supported_standards
        else:
            standards = []

        self.standards: List[str] = standards
        self._analyser = analyser
        self._filter_standards_names()
        self._validate_standards()
        self._check_other_implemented_standards()

    def _filter_standards_names(self):
        """
        Converts all the Neo standards names to upper kebab case
        """
        filtered_standards = set()
        for standard in self.standards:
            standard = standard.strip()
            if standard.upper().startswith('NEP'):
                rebab_case = re.sub(r'([A-Z]+|(.\d+)+|\d+)[^a-zA-Z\d]?', r'\g<1>-', standard.upper())
                if rebab_case.endswith('-'):
                    rebab_case = rebab_case[:-1]
                standard = rebab_case

            filtered_standards.add(standard)

        self.standards.clear()
        self.standards.extend(filtered_standards)

    def get_methods_by_display_name(self, method_id: str) -> List[Method]:
        methods = []
        for symbol_id, symbol in self.symbols.items():
            if isinstance(symbol, Method) and symbol.is_public:
                display_name = symbol.external_name if symbol.external_name is not None else symbol_id
                if display_name == method_id:
                    methods.append(symbol)
        return methods

    def _validate_standards(self):
        try:
            for standard in self.standards:
                if standard in supportedstandard.neo_standards:
                    standard_was_found = False
                    standard_errors = {}

                    standard_index = 0
                    while standard_index < len(supportedstandard.neo_standards[standard]) and not standard_was_found:
                        current_standard = supportedstandard.neo_standards[standard][standard_index]
                        current_standard_errors = []

                        # validate standard's methods
                        for standard_method in current_standard.methods:
                            method_id = standard_method.external_name
                            is_implemented = False

                            found_methods = self.get_methods_by_display_name(method_id)
                            for method in found_methods:
                                if isinstance(method, Method) and current_standard.match_definition(standard_method, method):
                                    is_implemented = True
                                    break

                            if not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard, method_id, standard_method)
                                )

                        # validate standard's events
                        events = [symbol for symbol in self.symbols.values() if isinstance(symbol, Event)]
                        # imported events should be included in the validation
                        for import_ in self._analyser.get_imports():
                            events.extend([event for event in import_.symbol_table.values()
                                           if isinstance(event, Event) and event not in events])

                        for standard_event in current_standard.events:
                            is_implemented = False
                            for event in events:
                                if (event.name == standard_event.name
                                        and current_standard.match_definition(standard_event, event)):
                                    is_implemented = True
                                    break

                            if not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard,
                                                                            standard_event.name,
                                                                            standard_event)
                                )

                        # validate optional methods
                        for optional_method in current_standard.optionals:
                            method_id = optional_method.external_name
                            is_implemented = False

                            found_methods = self.get_methods_by_display_name(method_id)
                            for method in found_methods:
                                if isinstance(method, Method) and current_standard.match_definition(optional_method, method):
                                    is_implemented = True
                                    break

                            if found_methods and not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard, method_id, optional_method)
                                )

                        standard_errors[standard_index] = current_standard_errors
                        standard_was_found = len(current_standard_errors) == 0
                        standard_index += 1

                    if not standard_was_found:
                        for error in standard_errors[0]:
                            self._log_error(error)

        except CompilerError.CompilerError:
            # stops the analyser if fail fast is activated
            pass

    def _check_other_implemented_standards(self):
        other_standards = supportedstandard.neo_standards.copy()
        # verify only standards that were not mentioned
        for standard in self.standards:
            if standard in supportedstandard.neo_standards:
                other_standards.pop(standard)

        # gets a list of all events
        events = [symbol for symbol in self.symbols.copy().values() if isinstance(symbol, Event)]
        for import_ in self._analyser.get_imports():
            events.extend([event for event in import_.symbol_table.values() if isinstance(event, Event)])

        # verify if the methods and events that were implemented corresponds to a standard
        for standard in other_standards:
            for current_standard in other_standards[standard]:
                # verify the methods
                methods_implemented = True
                standard_methods = current_standard.methods
                index = 0

                while methods_implemented and index < len(standard_methods):
                    standard_method = standard_methods[index]
                    method_id = standard_method.external_name
                    found_methods = self.get_methods_by_display_name(method_id)

                    methods_implemented = any(
                        current_standard.match_definition(standard_method, method) for method in found_methods
                    )
                    index += 1

                if not methods_implemented:
                    continue    # if even one of the methods was not implemented, then check the next standard

                # verify the events
                events_implemented = True
                standard_events = current_standard.events
                index = 0

                while events_implemented and index < len(standard_events):
                    standard_event = standard_events[index]
                    events_implemented = any(
                        (event.name == standard_event.name and
                         current_standard.match_definition(standard_event, event)) for event in events
                    )
                    index += 1

                if not events_implemented:
                    continue    # if even one of the events was not implemented, then check the next standard

                self.standards.append(standard)
                break



__all__ = [
    'Policy',
]


from boa3.builtin.type import UInt160


class Policy:
    """
    A class used to represent the Policy native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/Policy>`__
    to learn more about the Policy class.
    """

    hash: UInt160

    @classmethod
    def get_fee_per_byte(cls) -> int:
        """
        Gets the network fee per transaction byte.

        >>> Policy.get_fee_per_byte()
        1000

        :return: the network fee per transaction byte
        :rtype: int
        """
        pass

    @classmethod
    def get_exec_fee_factor(cls) -> int:
        """
        Gets the execution fee factor. This is a multiplier that can be adjusted by the committee to adjust the system fees
        for transactions.

        >>> Policy.get_exec_fee_factor()
        30

        :return: the execution fee factor
        :rtype: int
        """
        pass

    @classmethod
    def get_storage_price(cls) -> int:
        """
        Gets the storage price.

        >>> Policy.get_storage_price()
        100000

        :return: the snapshot used to read data
        :rtype: int
        """
        pass

    @classmethod
    def is_blocked(cls, account: UInt160) -> bool:
        """
        Determines whether the specified account is blocked.

        >>> Policy.is_blocked(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))
        False

        :param account: the account to be checked
        :type account: UInt160

        :return: whether the account is blocked or not
        :rtype: bool
        """
        pass



from __future__ import annotations

from typing import Any, Dict

from boa3.internal.model.builtin.interop.nativecontract import CryptoLibContract
from boa3.internal.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.internal.model.method import Method


class CryptoLibClass(INativeContractClass):
    """
    A class used to represent CryptoLib native contract
    """

    def __init__(self):
        super().__init__('CryptoLib', CryptoLibContract)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.internal.model.builtin.interop.interop import Interop

        if len(self._class_methods) == 0:
            self._class_methods = {
                'murmur32': Interop.Murmur32,
                'sha256': Interop.Sha256,
                'ripemd160': Interop.Ripemd160,
                'verify_with_ecdsa': Interop.VerifyWithECDsa,
                'bls12_381_add': Interop.Bls12381Add,
                'bls12_381_deserialize': Interop.Bls12381Deserialize,
                'bls12_381_equal': Interop.Bls12381Equal,
                'bls12_381_mul': Interop.Bls12381Mul,
                'bls12_381_pairing': Interop.Bls12381Pairing,
                'bls12_381_serialize': Interop.Bls12381Serialize,
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> CryptoLibClass:
        if value is None or cls._is_type_of(value):
            return _CryptoLib

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, CryptoLibClass)


_CryptoLib = CryptoLibClass()



import ast
from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StartsWithMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.internal.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            self_type = Type.bytes

        identifier = 'startswith'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'value': Variable(self_type),
            'start': Variable(Type.int),
            'end': Variable(Type.optional.build(Type.int)),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("{0}".format(Type.none.default_value)
                                ).body[0].value

        super().__init__(identifier, args, defaults=[start_default, end_default], return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # fix start and end indexes if they are negative or greater than len(string)
        code_generator.swap_reverse_stack_items(4)
        code_generator.swap_reverse_stack_items(3)
        code_generator.fix_negative_index()
        code_generator.swap_reverse_stack_items(2)

        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(None)
        if_end_is_none = code_generator.convert_begin_if()
        code_generator.change_jump(if_end_is_none, Opcode.JMPIFNOT)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        else_end_is_not_none = code_generator.convert_begin_else(if_end_is_none, is_internal=True)
        code_generator.fix_negative_index()
        code_generator.convert_end_if(else_end_is_not_none, is_internal=True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.fix_index_out_of_range(True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.fix_index_out_of_range(True)

        # if start < len(string):
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        if_start_lt_string_len = code_generator.convert_begin_if()
        code_generator.change_jump(if_start_lt_string_len, Opcode.JMPGE)

        #   string = string[start:end]
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.convert_get_substring(is_internal=True)

        #   if len(string) <= len(substring):
        code_generator.convert_literal(0)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        #       return string[0:len(substring)] == substring
        if_str_is_not_smaller_than_substr = code_generator.convert_begin_if()
        code_generator.change_jump(if_str_is_not_smaller_than_substr, Opcode.JMPGT)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)

        #   elif len(string) > len(substring):
        else_str_is_smaller = code_generator.convert_begin_else(if_str_is_not_smaller_than_substr, is_internal=True)

        for _ in self.args:
            code_generator.remove_stack_top_item()
        #   return False
        code_generator.convert_literal(False)

        code_generator.convert_end_if(else_str_is_smaller, is_internal=True)

        # elif start >= len(string):
        else_start_gt_string_len = code_generator.convert_begin_else(if_start_lt_string_len, is_internal=True)
        for _ in self.args:
            code_generator.remove_stack_top_item()
        #   return False
        code_generator.convert_literal(False)

        code_generator.convert_end_if(else_start_gt_string_len, is_internal=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return StartsWithMethod(value)
        return super().build(value)



import ast
from typing import Dict

from boa3.internal.model.builtin.method.listmethod import ListMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class ListGenericMethod(ListMethod):

    def __init__(self, value: IType = None):
        from boa3.internal.model.type.type import Type

        if value is None:
            value = Type.any

        args: Dict[str, Variable] = {
            'value': Variable(value),
        }

        value_default = ast.parse("{0}".format(Type.sequence.default_value)
                                  ).body[0].value

        return_value = Type.any if value is Type.any else []

        if return_value is not Type.any:
            for type_ in value.union_types:
                if type_ is Type.str:
                    return_value.append(Type.str)
                elif type_ is Type.bytes:
                    return_value.append(Type.int)
                elif Type.sequence.is_type_of(type_):
                    return_value.append(type_.value_type)
                elif Type.mapping.is_type_of(type_):
                    return_value.append(type_.key_type)

            if Type.str in value.union_types and Type.bytes in value.union_types:
                return_value.remove(Type.str)

        return_type = Type.list.build_collection(return_value)

        super().__init__(args, return_type, [value_default])

    def generate_pack_opcodes(self, code_generator):
        from boa3.internal.model.builtin.method import ListBytesStringMethod
        from boa3.internal.model.type.annotation.uniontype import UnionType
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        error_message = 'Invalid value given, it should be an iterable'

        # if instance(value, str)
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.ByteString)
        is_byte_string = code_generator.convert_begin_if()

        #   byte string implementation
        (ListBytesStringMethod(Type.str)
         if (isinstance(self._arg_value.type, UnionType) and
             Type.str in self._arg_value.type.union_types and
             Type.bytes not in self._arg_value.type.union_types)
         else ListBytesStringMethod()
         ).generate_pack_opcodes(code_generator)

        is_byte_string = code_generator.convert_begin_else(is_byte_string, is_internal=True)
        # elif isinstance(value, list):
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Array)
        is_list = code_generator.convert_begin_if()

        #   list implementation
        from boa3.internal.model.builtin.method import ListSequenceMethod
        ListSequenceMethod().generate_pack_opcodes(code_generator)

        is_list = code_generator.convert_begin_else(is_list, is_internal=True)
        # elif isinstance(value, map):
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Map)
        is_map = code_generator.convert_begin_if()

        #   map implementation
        from boa3.internal.model.builtin.method import ListMappingMethod
        ListMappingMethod().generate_pack_opcodes(code_generator)

        is_map = code_generator.convert_begin_else(is_map, is_internal=True)
        # else:
        code_generator.convert_literal(error_message)
        # raise error
        code_generator.convert_raise_exception()

        code_generator.convert_end_if(is_map, is_internal=True)
        code_generator.convert_end_if(is_list, is_internal=True)
        code_generator.convert_end_if(is_byte_string, is_internal=True)



from typing import Any

from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.internal.model.type.itype import IType


class PublicKeyType(ECPointType):
    """
    A class used to indicate that a parameter or return on the manifest is a PublicKey. It's a subclass of ECPointType.
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'PublicKey'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _PublicKey


_PublicKey = PublicKeyType()



__all__ = [
    'Nep11TransferEvent',
    'Nep17TransferEvent',
    'NeoAccountState',
    'abort',
    'to_hex_str',
    'to_script_hash',
]

from typing import Any, Optional, Union

from boa3.builtin.compile_time import CreateNewEvent
from boa3.builtin.type import ECPoint, UInt160, Event

Nep11TransferEvent: Event = CreateNewEvent(
    [
        ('from', Union[UInt160, None]),
        ('to', Union[UInt160, None]),
        ('amount', int),
        ('tokenId', Union[str, bytes])
    ],
    'Transfer'
)
"""
The NEP-11 Transfer event that should be triggered whenever a non-fungible token is transferred, minted or burned. It 
needs the addresses of the sender, receiver, amount transferred and the id of the token.

Check out the `proposal <https://github.com/neo-project/proposals/blob/master/nep-11.mediawiki>`__ or 
`Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/nep11>`__ about this NEP.

>>> Nep11TransferEvent(b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',  # when calling, it will return None, but the event will be triggered
...                    b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0', 1, '01')
{
    'name': 'Transfer',
    'script hash': b'\\x13\\xb4\\x51\\xa2\\x1c\\x10\\x12\\xd6\\x13\\x12\\x19\\x0c\\x15\\x61\\x9b\\x1b\\xd1\\xa2\\xf4\\xb2',
    'state': {
        'from': b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',
        'to': b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0',
        'amount': 1,
        'tokenId': '01'
    }
}

:meta hide-value:
"""


Nep17TransferEvent: Event = CreateNewEvent(
    [
        ('from', Union[UInt160, None]),
        ('to', Union[UInt160, None]),
        ('amount', int)
    ],
    'Transfer'
)
"""
The NEP-17 Transfer event that should be triggered whenever a fungible token is transferred, minted or burned. It needs
the addresses of the sender, receiver and the amount transferred.

Check out the `proposal <https://github.com/neo-project/proposals/blob/master/nep-17.mediawiki>`__ or 
`Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/nep17>`__ about this NEP.

>>> Nep17TransferEvent(b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',  # when calling, it will return None, but the event will be triggered
...                    b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0', 100)
{
    'name': 'Transfer',
    'script hash': b'\\x17\\xe3\\xca\\x91\\xca\\xb7\\xaf\\xdd\\xe6\\xba\\x07\\xaa\\xba\\xa1\\x66\\xab\\xcf\\x00\\x04\\x50',
    'state': {
        'from': b'\\xd1\\x17\\x92\\x82\\x12\\xc6\\xbe\\xfa\\x05\\xa0\\x23\\x07\\xa1\\x12\\x55\\x41\\x06\\x55\\x10\\xe6',
        'to': b'\\x18\\xb7\\x30\\x14\\xdf\\xcb\\xee\\x01\\x30\\x00\\x13\\x9b\\x8d\\xa0\\x13\\xfb\\x96\\xac\\xd1\\xc0',
        'amount': 100
    }
}

:meta hide-value:
"""


def abort(msg: Optional[str] = None):
    """
    Aborts the execution of a smart contract. Using this will cancel the changes made on the blockchain by the
    transaction.

    >>> abort()     # abort doesn't return anything by itself, but the execution will stop and the VMState will be FAULT
    VMState.FAULT

    >>> abort('abort message')
    VMState.FAULT

    """
    pass


class NeoAccountState:
    """
    Represents the account state of NEO token in the NEO system.

    :ivar balance: the current account balance, which equals to the votes cast
    :vartype balance: int
    :ivar height: the height of the block where the balance changed last time
    :vartype height: int
    :ivar vote_to: the voting target of the account
    :vartype vote_to: ECPoint
    """

    def __init__(self):
        from boa3.internal import constants
        self.balance: int = 0
        self.height: int = 0
        self.vote_to: ECPoint = ECPoint(bytes(constants.SIZE_OF_ECPOINT))


def to_script_hash(data_bytes: Any) -> bytes:
    """
    Converts a data to a script hash.

    >>> to_script_hash(ECPoint(bytes(range(33))))
    b'\\x12\\xc8z\\xfb3k\\x1e4>\\xb3\\x83\\tK\\xc7\\xdch\\xe5\\xee\\xc7\\x98'

    >>> to_script_hash(b'1234567891')
    b'\\x4b\\x56\\x34\\x17\\xed\\x99\\x7f\\x13\\x22\\x67\\x40\\x79\\x36\\x8b\\xa2\\xcd\\x72\\x41\\x25\\x6d'

    :param data_bytes: data to hash
    :type data_bytes: Any
    :return: the script hash of the data
    :rtype: bytes
    """
    pass


def to_hex_str(data: bytes) -> str:
    """
    Converts bytes into its string hex representation.

    >>> to_hex_str(ECPoint(bytes(range(33))))
    '201f1e1d1c1b1a191817161514131211100f0e0d0c0b0a09080706050403020100'

    >>> to_hex_str(b'1234567891')
    '31393837363534333231'

    :param data: data to represent as hex.
    :type data: bytearray or bytes

    :return: the hex representation of the data
    :rtype: str
    """
    pass



__all__ = [
    'BoaTestRunner'
]

import os.path
import threading
from typing import Callable, List, Sequence, Optional, Tuple, Union

from boa3.internal import env
from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.smart_contract.contractcollection import ContractCollection
from boa3_test.test_drive.neoxp.model.neoxpconfig import NeoExpressConfig
from boa3_test.test_drive.testrunner.blockchain.block import TestRunnerBlock as Block
from boa3_test.test_drive.testrunner.neo_test_runner import NeoTestRunner
from boa3_test.tests.test_drive.neoxp import utils as boa_neoxp_utils
from boa3_test.tests.test_drive.testrunner.boa_neoxp_batch import BoaNeoExpressBatch


class BoaTestRunner(NeoTestRunner):
    _DEFAULT_ACCOUNT = boa_neoxp_utils.get_default_account()

    def __init__(self, neoxp_path: str = None, runner_id: str = None, cleanup_files: bool = True):
        if not isinstance(neoxp_path, str):
            neoxp_path = os.path.join(env.NEO_EXPRESS_INSTANCE_DIRECTORY, 'default.neo-express')

        super().__init__(neoxp_path, runner_id)
        self._batch = BoaNeoExpressBatch(self._neoxp_config)

        self._clear_files_when_destroyed = cleanup_files

    def _set_up_neoxp_config(self) -> NeoExpressConfig:
        default_config = boa_neoxp_utils._NEOXP_CONFIG
        if self._neoxp_abs_path == os.path.abspath(default_config.config_path):
            return default_config
        return super()._set_up_neoxp_config()

    def _set_up_generate_file_names(self, file_name: str):
        from boa3_test.test_drive import utils
        runner_specific_id = utils.create_custom_id(file_name)
        super()._set_up_generate_file_names(runner_specific_id)

    def add_neo(self, script_hash_or_address: Union[bytes, str], amount: int):
        address = boa_neoxp_utils.get_account_from_script_hash_or_id(script_hash_or_address)
        self._batch.transfer_assets(sender=self._DEFAULT_ACCOUNT, receiver=address,
                                    quantity=amount,
                                    asset='NEO')

    def add_gas(self, script_hash_or_address: Union[bytes, str], amount: int):
        address = boa_neoxp_utils.get_account_from_script_hash_or_id(script_hash_or_address)
        gas_decimals = 8
        self._batch.transfer_assets(sender=self._DEFAULT_ACCOUNT, receiver=address,
                                    asset='GAS', decimals=gas_decimals,
                                    quantity=(amount / (10 ** gas_decimals)))

    def _internal_generate_files(self, methods_to_call: List[Tuple[Callable, Sequence]]):
        worker_threads = []

        if len(methods_to_call) > 0:
            last_method, last_method_args = methods_to_call[-1]

            for method, args in methods_to_call[:-1]:
                work_thread = threading.Thread(target=method, args=args)
                worker_threads.append(
                    work_thread
                )
                work_thread.start()

            last_method(*last_method_args)

        for worker in worker_threads:
            worker.join()

    def _get_genesis_block(self) -> Optional[Block]:
        return boa_neoxp_utils.get_genesis_block()

    def _get_block(self, block_hash_or_index) -> Optional[Block]:
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return boa_neoxp_utils.get_block(self._neoxp_abs_path, block_hash_or_index,
                                         check_point_file=check_point_path)

    def _get_tx(self, tx_hash: UInt256):
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return boa_neoxp_utils.get_transaction(self._neoxp_abs_path, tx_hash,
                                               check_point_file=check_point_path)

    def _get_tx_log(self, tx_hash: UInt256, contract_collection: ContractCollection):
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return boa_neoxp_utils.get_transaction_log(self._neoxp_abs_path, tx_hash,
                                                   check_point_file=check_point_path,
                                                   contract_collection=contract_collection)

    def _get_oracle_resp(self, url: str, response_path: str, request_id: int = None) -> List[UInt256]:
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return boa_neoxp_utils.oracle_response(self._neoxp_abs_path, url, response_path, request_id,
                                               check_point_file=check_point_path)

    def __del__(self):
        self.reset()
        if self._clear_files_when_destroyed:
            paths_to_delete = [
                self.get_full_path(self._CHECKPOINT_FILE),
                self.get_full_path(self._BATCH_FILE),
                self.get_full_path(self._INVOKE_FILE)
            ]
            for path in paths_to_delete:
                try:
                    if os.path.exists(path):
                        os.remove(path)
                except BaseException as e:
                    print(e)
                    continue



from typing import Dict

from boa3.internal.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.internal.model.variable import Variable


class Bls12381SerializeMethod(CryptoLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'bls12_381_serialize'
        native_identifier = 'bls12381Serialize'
        args: Dict[str, Variable] = {
            'g': Variable(Type.any),
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.bytes)



from typing import Dict

from boa3.internal.model.symbol import ISymbol


class ImportData:
    def __init__(self, import_symbol: ISymbol, origin_id: str, origin_file: str = None):
        self._imported_symbol = import_symbol
        self._origin_id = origin_id
        self._origin_file = origin_file

    @property
    def origin(self) -> str:
        if self._origin_id is not None:
            return self._origin_id

        if hasattr(self._imported_symbol, 'origin'):
            return self._imported_symbol.origin

        return None

    @property
    def origin_file(self) -> str:
        return self._origin_file

    @property
    def all_symbols(self) -> Dict[str, ISymbol]:
        if hasattr(self._imported_symbol, 'all_symbols'):
            return self._imported_symbol.all_symbols

        if hasattr(self._imported_symbol, 'symbols'):
            return self._imported_symbol.symbols

        return {}



from typing import Dict

from boa3.internal.model.builtin.method.builtinevent import IBuiltinEvent
from boa3.internal.model.variable import Variable


class Nep11TransferEvent(IBuiltinEvent):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'Nep11TransferEvent'
        type_token_id = Type.union.build([Type.str, Type.bytes])

        args: Dict[str, Variable] = {
            'from_addr': Variable(Type.union.build([Type.none, UInt160Type.build()])),
            'to_addr': Variable(Type.union.build([Type.none, UInt160Type.build()])),
            'amount': Variable(Type.int),
            'tokenId': Variable(type_token_id),
        }
        super().__init__(identifier, args)
        self.name = 'Transfer'



import json
import os.path
from typing import List

from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.model.neoxpaccount import NeoExpressAccount

_VERSION = utils.DEFAULT_ACCOUNT_VERSION  # default Neo account version


class NeoExpressConfig:
    def __init__(self, neo_express_path: str):
        if not isinstance(neo_express_path, str):
            raise TypeError(f"Invalid data type {type(neo_express_path)}. Expecting str")
        elif not os.path.isfile(neo_express_path):
            raise TypeError(f"Given path doesn't exist '{neo_express_path}'")
        elif not neo_express_path.endswith('.neo-express'):
            raise TypeError(f"Given file is not a valid neo express config file '{neo_express_path}'")

        with open(neo_express_path, 'rb') as config_file:
            config_json = json.loads(config_file.read())

        self._neo_express_config_path = neo_express_path
        self._magic = config_json['magic']
        self._version = config_json['address-version']

        NeoExpressConfig._VERSION = self._version  # avoid circular import to set accounts

        genesis_account = None
        if len(config_json['consensus-nodes']) > 0:
            node1_accounts = _wallet_accounts_from_json(config_json['consensus-nodes'][0]['wallet'])
            genesis_account = next((account for account in node1_accounts if account.name is None),
                                   None)
            if genesis_account is not None:
                genesis_account._name = 'genesis'
            accounts = node1_accounts
        else:
            accounts = []

        for consensus in config_json['consensus-nodes'][1:]:
            accounts.extend(_wallet_accounts_from_json(consensus['wallet']))

        for custom_wallet in config_json['wallets']:
            accounts.extend(_wallet_accounts_from_json(custom_wallet))
        self._accounts = accounts
        self._default_account = (genesis_account
                                 if genesis_account is not None or len(accounts) == 0
                                 else accounts[0])
        self._genesis_block = None

    @property
    def magic(self):
        return self._magic

    @property
    def version(self):
        return self._version

    @property
    def accounts(self):
        return self._accounts.copy()

    @property
    def default_account(self):
        return self._default_account

    @property
    def genesis_block(self):
        return self._genesis_block

    @property
    def config_path(self):
        return self._neo_express_config_path


def _wallet_accounts_from_json(wallet_json: dict) -> List[Account]:
    wallet_accounts = []
    name = wallet_json['name']
    default_was_set = False

    for json_account in wallet_json['accounts']:
        account = NeoExpressAccount.from_json(json_account)
        if not isinstance(account, Account):
            continue

        wallet_accounts.append(account)
        account._version = _VERSION
        if not default_was_set and hasattr(json_account, 'is-default') and json_account['is-default']:
            account._name = name

    if len(wallet_accounts) > 0 and not default_was_set:
        wallet_accounts[0]._name = name

    return wallet_accounts



from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintClassMethod(PrintMethod):

    def __init__(self, arg_value: IType):
        if not isinstance(arg_value, UserClass):
            from boa3.internal.model.type.classes import userclass
            arg_value = userclass._EMPTY_CLASS

        super().__init__(arg_value)
        self._arg_type = arg_value

    def _generate_print_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.type.type import Type

        # class to dict, then print as json
        code_generator.insert_opcode(Opcode.UNPACK, add_to_stack=[Type.any, Type.int])
        code_generator.remove_stack_top_item()
        code_generator.convert_new_map(Type.dict)

        # add each value to the dict
        for variable_name in self._arg_type.instance_variables:
            code_generator.insert_opcode(Opcode.TUCK, add_to_stack=[Type.any])
            code_generator.convert_literal(variable_name)
            item_address = code_generator.bytecode_size
            code_generator.swap_reverse_stack_items(3, rotate=True)
            code_generator.convert_set_item(item_address, index_inserted_internally=True)

        # print as json
        code_generator.convert_builtin_method_call(Interop.JsonSerialize, is_internal=True)



from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IsDigitMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.internal.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            self_type = Type.bytes

        identifier = 'isdigit'
        args: Dict[str, Variable] = {'self': Variable(self_type)}

        super().__init__(identifier, args, return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # index = len(string) - 1
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        # isdigit = True
        code_generator.convert_literal(True)

        # if len(string) > 0:
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(-1)
        if_string_not_empty = code_generator.convert_begin_if()
        code_generator.change_jump(if_string_not_empty, Opcode.JMPEQ)

        #   while index >= 0:
        while_str_is_analysed = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal('0')

        #       if ord(string[index]) < ord('0'):
        if_char_lt_0 = code_generator.convert_begin_if()
        code_generator.change_jump(if_char_lt_0, Opcode.JMPGE)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #           return False
        code_generator.convert_literal(False)
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_char_lt_0, is_internal=True)

        #       elif ord(string[index]) > ord('9'):
        code_generator.convert_literal('9')
        if_char_gt_9 = code_generator.convert_begin_if()
        code_generator.change_jump(if_char_gt_9, Opcode.JMPLE)
        #           return False
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(False)
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_char_gt_9, is_internal=True)

        #       index -= 1
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.swap_reverse_stack_items(2)

        # verify if already passed through the whole string
        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)
        code_generator.convert_end_while(while_str_is_analysed, while_condition, is_internal=True)

        # elif len(string) == 0:
        else_string_is_empty = code_generator.convert_begin_else(if_string_not_empty, is_internal=True)
        code_generator.remove_stack_top_item()
        #   return False
        code_generator.convert_literal(False)
        code_generator.convert_end_if(else_string_is_empty, is_internal=True)

        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return IsDigitMethod(value)
        return super().build(value)



from typing import Dict

from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.builtin.interop.iterator.iteratortype import IteratorType
from boa3.internal.model.variable import Variable


class GetIteratorValue(InteropMethod):
    def __init__(self, iterator: IteratorType):
        syscall = 'System.Iterator.Value'
        identifier = '-get_iterator_value'
        args: Dict[str, Variable] = {'self': Variable(iterator)}

        result_type = iterator.item_type
        from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
        if isinstance(iterator._origin_collection, MappingType):
            from boa3.internal.model.type.type import Type
            result_type = Type.tuple.build((iterator.key_type, iterator.item_type))
        super().__init__(identifier, syscall, args, return_type=result_type)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def self_arg(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        super().generate_internal_opcodes(code_generator)

        # if result is not Struct
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Struct)
        is_struct = code_generator.convert_begin_if()

        #   cast(result, list)
        code_generator.convert_cast(Type.list, is_internal=True)

        code_generator.convert_end_if(is_struct, is_internal=True)


class IteratorValueProperty(IBuiltinProperty):
    def __init__(self, iterator: IteratorType):
        identifier = 'iterator_value'
        getter = GetIteratorValue(iterator)
        super().__init__(identifier, getter)



import ast
from typing import Dict, Optional

from boa3.internal.model.builtin.classmethod.indexmethod import IndexMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.model.type.primitive.strtype import StrType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IndexBytesStringMethod(IndexMethod):

    def __init__(self, self_type: Optional[StrType] = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(self_type, (StrType, BytesType)):
            self_type = Type.str

        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'x': Variable(self_type),
            'start': Variable(Type.int),
            'end': Variable(Type.optional.build(Type.int)),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("{0}".format(Type.none.default_value)
                                ).body[0].value

        super().__init__(args, defaults=[start_default, end_default])

    def validate_parameters(self, *params: IExpression) -> bool:
        if 2 <= len(params) <= 4:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        self_type: IType = params[0].type

        if not isinstance(self_type, (StrType, BytesType)):
            return False

        if not self_type.is_type_of(params[1]):
            return False

        return True

    @property
    def error_message(self) -> str:
        return 'substring not found' if isinstance(self._arg_self.type, StrType) else 'subsequence of bytes not found'

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # fix possible negative or too big indexes for start and end parameters
        code_generator.swap_reverse_stack_items(2)
        code_generator.swap_reverse_stack_items(4)
        code_generator.swap_reverse_stack_items(2)
        code_generator.fix_negative_index()
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(None)
        if_end_is_none = code_generator.convert_begin_if()
        code_generator.change_jump(if_end_is_none, Opcode.JMPIFNOT)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        else_end_is_not_none = code_generator.convert_begin_else(if_end_is_none, is_internal=True)
        code_generator.fix_negative_index()
        code_generator.convert_end_if(else_end_is_not_none, is_internal=True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.fix_index_out_of_range(True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.fix_index_out_of_range(True)
        code_generator.swap_reverse_stack_items(2)

        # change order of items on stack
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)

        # current_index = start
        # while (size < end - current_index):
        while_start = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(5)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.duplicate_stack_item(6)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        #   if str[index:current_index + size] == substr:
        if_was_found = code_generator.convert_begin_if()
        code_generator.change_jump(if_was_found, Opcode.JMPIFNOT)
        #       break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_was_found)
        #   current_index += 1
        code_generator.insert_opcode(Opcode.INC)

        condition_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)
        code_generator.convert_end_while(while_start, condition_address, is_internal=True)

        # if current_index + size >= end:
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.duplicate_stack_item(4)
        check_was_found = code_generator.convert_begin_if()
        code_generator.change_jump(check_was_found, Opcode.JMPGE)
        #   raise Exception('substring not found')
        code_generator.convert_literal(self.error_message)
        code_generator.convert_raise_exception()
        code_generator.convert_end_if(check_was_found)
        # else:
        #   return current_index
        for _ in range(1, len(code_generator._stack)):
            code_generator.remove_stack_item(2)



from typing import Dict

from boa3.internal.model.builtin.interop.nativecontract import Nep17Method
from boa3.internal.model.variable import Variable


class BalanceOfMethod(Nep17Method):

    def __init__(self, contract_script_hash: bytes):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'balanceOf'
        native_identifier = 'balanceOf'
        args: Dict[str, Variable] = {'account': Variable(UInt160Type.build())}
        super().__init__(identifier, native_identifier, args, return_type=Type.int, script_hash=contract_script_hash)



from typing import Any, List, Optional, Tuple

from boa3.internal import constants
from boa3.internal.model.method import Method
from boa3.internal.model.type.classes.classtype import ClassType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.neo.vm.type.AbiType import AbiType


class UInt256Type(BytesType, ClassType):
    """
    A class used to represent Neo's UInt256 type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'UInt256'
        from boa3.internal.model.builtin.method.uint256method import UInt256Method
        self._constructor = UInt256Method(self)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Hash256

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @property
    def default_value(self) -> Any:
        return bytes(constants.SIZE_OF_INT256)

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _UInt256

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, UInt256Type)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).is_instance_opcodes()

    def generate_is_instance_type_check(self, code_generator):
        from boa3.internal.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).generate_is_instance_type_check(code_generator)

    def _generate_specific_class_type_check(self, code_generator) -> List[int]:
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(constants.SIZE_OF_INT256)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        return []

    def _is_instance_inner_opcodes(self, jmp_to_if_false: int = 0) -> List[Tuple[Opcode, bytes]]:
        push_int_opcode, size_data = OpcodeHelper.get_push_and_data(constants.SIZE_OF_INT256)

        return [
            (Opcode.SIZE, b''),  # return len(value) == 32
            (push_int_opcode, size_data),
            (Opcode.NUMEQUAL, b'')
        ]


_UInt256 = UInt256Type()



import ast
from typing import Dict

from boa3.internal.model.builtin.method.intmethod import IntMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IntByteStringMethod(IntMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type

        args: Dict[str, Variable] = {
            'value': Variable(Type.union.build([
                Type.bytes,
                Type.str
            ])),
            'base': Variable(Type.int)
        }

        value_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value

        base_default = ast.parse("{0}".format(10)
                                 ).body[0].value

        super().__init__(args, [value_default, base_default])

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type

        # if base >= 37:
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(37)
        if_base_ge_37 = code_generator.convert_begin_if()
        code_generator.change_jump(if_base_ge_37, Opcode.JMPLT)
        #   assert False
        code_generator.convert_literal(False)
        code_generator.convert_assert()
        code_generator.convert_end_if(if_base_ge_37, is_internal=True)

        # if base == 1:
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        if_base_eq_1 = code_generator.convert_begin_if()
        code_generator.change_jump(if_base_eq_1, Opcode.JMPNE)
        #   assert False
        code_generator.convert_literal(False)
        code_generator.convert_assert()
        code_generator.convert_end_if(if_base_eq_1, is_internal=True)

        # if base <= -1:
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(-1)
        if_base_le_m1 = code_generator.convert_begin_if()
        code_generator.change_jump(if_base_le_m1, Opcode.JMPGT)
        #   assert False
        code_generator.convert_literal(False)
        code_generator.convert_assert()
        code_generator.convert_end_if(if_base_le_m1, is_internal=True)

        # if value[0] != 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.convert_literal('0')
        if_first_char_ne_0 = code_generator.convert_begin_if()
        code_generator.change_jump(if_first_char_ne_0, Opcode.JMPNE)

        #   base_hint_char = value[1]
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(1)
        code_generator.convert_get_item(index_inserted_internally=True)

        #   base_found = False
        code_generator.convert_literal(False)

        #   # verifies which base user might be using and change to a cleaner version, e.g., ('0b101', 0) becomes ('101', 2)
        self._verify_is_specific_base('b', 2, code_generator)
        self._verify_is_specific_base('B', 2, code_generator)
        self._verify_is_specific_base('o', 8, code_generator)
        self._verify_is_specific_base('O', 8, code_generator)
        self._verify_is_specific_base('x', 16, code_generator)
        self._verify_is_specific_base('X', 16, code_generator)

        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

        code_generator.convert_end_if(if_first_char_ne_0, is_internal=True)

        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        # index = len(value) - 1
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)
        code_generator.convert_assert()
        # sum = 0
        code_generator.convert_literal(0)
        # mult = 1
        code_generator.convert_literal(1)

        # while (index >=  0):
        while_start = code_generator.convert_begin_while()

        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.convert_cast(Type.int, is_internal=True)

        #   if ord(value[index]) >= ord('a') :
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(ord('a'))
        if_char_is_lower = code_generator.convert_begin_if()
        code_generator.change_jump(if_char_is_lower, Opcode.JMPLT)

        #       assert ord(value[index]) <= base + 86
        # # the number 86 is being used because adding it to the base will return the valid characters used on that base
        # # e.g., base 11 lets you use the letter 'a', so, 86 + 11 = 97 == ord('a')
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(7)
        code_generator.convert_literal(86)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.convert_operation(BinaryOp.LtE, is_internal=True)
        code_generator.convert_assert()

        #       sum = sum + (ord(value[index]) - 87) * mult
        code_generator.convert_literal(87)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Mul, is_internal=True)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()

        #   elif ord(value[index]) >= ord('A'):
        else_char_is_not_lower = code_generator.convert_begin_else(if_char_is_lower, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(ord('A'))
        if_char_is_upper = code_generator.convert_begin_if()
        code_generator.change_jump(if_char_is_upper, Opcode.JMPLT)

        #       assert ord(value[index]) <= base + 54
        # # the number 54 is being used because adding it to the base will return the valid characters used on that base
        # # e.g., base 11 lets you use the letter 'A', so, 54 + 11 = 65 == ord('A')
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(7)
        code_generator.convert_literal(54)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.convert_operation(BinaryOp.LtE, is_internal=True)
        code_generator.convert_assert()

        #       sum = sum + (ord(value[index]) - 55) * mult
        code_generator.convert_literal(55)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Mul, is_internal=True)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()

        #   else:
        #       assert ord(value[index]) >= ord('0')
        else_char_is_not_upper = code_generator.convert_begin_else(if_char_is_upper, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(ord('0'))
        code_generator.convert_operation(BinaryOp.GtE)
        code_generator.convert_assert()

        #       assert ord(value[index]) < base + ord('0')
        # # if base is equal or less than 10, the char need to be ord('0') <= ord(char) < ord(str(base))
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(7)
        code_generator.convert_literal(ord('0'))
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.convert_operation(BinaryOp.LtE, is_internal=True)
        code_generator.convert_assert()

        #       sum = sum + (ord(value[index]) - ord('0')) * mult
        code_generator.convert_literal(ord('0'))
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Mul, is_internal=True)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()

        code_generator.convert_end_if(else_char_is_not_lower, is_internal=True)
        code_generator.convert_end_if(else_char_is_not_upper, is_internal=True)
        code_generator.swap_reverse_stack_items(3)
        #   index -= 1
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.swap_reverse_stack_items(3)
        code_generator.duplicate_stack_item(5)
        #   mult = mult * base
        code_generator.convert_operation(BinaryOp.Mul, is_internal=True)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)
        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # return sum
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

    def _verify_is_specific_base(self, char: str, corresponding_base: int, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp

        # if base_found:
        if_base_was_found = code_generator.convert_begin_if()
        #   pass
        code_generator.convert_literal(True)

        # elif not base_found:
        else_base_was_not_found = code_generator.convert_begin_else(if_base_was_found, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(char)

        #   if char != value[1]:
        if_chars_not_equal = code_generator.convert_begin_if()
        code_generator.change_jump(if_chars_not_equal, Opcode.JMPEQ)
        #       pass
        code_generator.convert_literal(False)

        #   elif char == value[1]:
        else_chars_equal = code_generator.convert_begin_else(if_chars_not_equal, is_internal=True)

        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(corresponding_base)
        code_generator.convert_operation(BinaryOp.Eq, is_internal=True)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Eq, is_internal=True)
        code_generator.convert_operation(BinaryOp.Xor, is_internal=True)
        #       if base == corresponding_base ^ base == 0:   # remove base indicator from ByteString and change to correct base if base is valid
        if_base_was_valid = code_generator.convert_begin_if()

        #           if base == 0:
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(0)
        if_base_is_0 = code_generator.convert_begin_if()
        code_generator.change_jump(if_base_is_0, Opcode.JMPNE)

        #               base = corresponding_base   # e.g., ('0b101', 0) becomes ('0b101', 2)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(corresponding_base)
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_end_if(if_base_is_0, is_internal=True)

        #           value = value[2:]       # e.g., ('0b101', 2) becomes ('101', 2)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_literal(2)
        code_generator.convert_get_sequence_ending()
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_end_if(if_base_was_valid)

        #       base_found = True
        code_generator.convert_literal(True)
        code_generator.convert_end_if(else_chars_equal)

        code_generator.convert_end_if(else_base_was_not_found)



from typing import List, Optional, Tuple

from boa3.internal.model.builtin.classmethod.countsequencemethod import CountSequenceMethod
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType


class CountSequenceGenericMethod(CountSequenceMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        super().__init__(sequence_type, arg_value)

    def _generic_verification(self, code_generator) -> Tuple[List[int], List[int]]:
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        jmps_to_inc, jmps_to_condition = super()._generic_verification(code_generator)

        # if value is Array
        code_generator.duplicate_stack_item(4)
        code_generator.insert_type_check(Type.sequence.stack_item)
        is_value_array = code_generator.convert_begin_if()

        #   item = sequence[index]
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)

        #   if item is Array
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(Type.sequence.stack_item)
        is_item_array = code_generator.convert_begin_if()

        #       if len(item) == len(value):
        code_generator.duplicate_stack_item(5)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        is_same_size = code_generator.convert_begin_if()
        code_generator.change_jump(is_same_size, Opcode.JMPNE)

        #           inner_index = len(item)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        #           while inner_index > 0
        while_start = code_generator.convert_begin_while()

        #               inner_index -= 1
        code_generator.insert_opcode(Opcode.DEC)

        #               if item[inner_index] != value[inner_index]
        code_generator.duplicate_stack_item(6)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.convert_operation(BinaryOp.NotEq)

        are_values_different = code_generator.convert_begin_if()
        code_generator.convert_loop_break()
        #                   break

        code_generator.convert_end_if(are_values_different)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # clean stack
        code_generator.remove_stack_item(2)

        #           if inner_index == 0:
        code_generator.convert_literal(0)
        is_equal = code_generator.convert_begin_if()
        code_generator.change_jump(is_equal, Opcode.JMPEQ)

        jmps_to_inc.append(is_equal)
        #               count += 1

        jmp_to_clean_stack = code_generator.convert_begin_if()
        code_generator.change_jump(jmp_to_clean_stack, Opcode.JMP)

        code_generator.convert_end_if(is_same_size, is_internal=True)
        code_generator.convert_end_if(is_item_array)
        code_generator.remove_stack_top_item()

        code_generator.convert_end_if(jmp_to_clean_stack)

        is_value_array = code_generator.convert_begin_else(is_value_array, is_internal=True)
        jmps_to_condition.append(is_value_array)

        return jmps_to_inc, jmps_to_condition



from typing import Any, List, Optional, Type

from boa3.internal.neo3.vm import VMState
from boa3_test.test_drive.model.invoker.neoinvoke import NeoInvoke
from boa3_test.test_drive.model.invoker.neoinvokeresult import NeoInvokeResult
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract
from boa3_test.test_drive.model.wallet.account import Account

__all__ = ['NeoInvokeCollection']


class NeoInvokeCollection:
    def __init__(self):
        self._internal_list: List[NeoInvoke] = []
        self._invoke_results: List[NeoInvokeResult] = []
        self._pending_results: List[NeoInvokeResult] = []

    def create_contract_invoke(self, contract: TestContract, operation: str, *args: Any) -> NeoInvoke:
        invoker = NeoInvoke(contract.name, operation, *args)
        invoker._contract = contract
        return invoker

    def append_contract_invoke(self, contract: TestContract, operation: str, *args: Any,
                               expected_result_type: Type = None) -> NeoInvokeResult:
        invoker = NeoInvoke(contract.name, operation, *args)
        invoker._contract = contract
        invoke_result = NeoInvokeResult(invoker, expected_result_type=expected_result_type)
        self._internal_list.append(invoker)
        self._invoke_results.append(invoke_result)
        self._pending_results.append(invoke_result)
        return invoke_result

    def clear(self, *, state: VMState = VMState.HALT):
        pending_results = self._pending_results.copy()
        self._invoke_results.clear()
        self._pending_results.clear()

        if state != VMState.HALT:
            for pending_invoke in pending_results:
                pending_invoke.cancel()
        return self._internal_list.clear()

    def to_json(self):
        if len(self._internal_list) > 0:
            return [invoke.to_json() for invoke in self._internal_list]

        return [{}]

    def set_results(self, new_result_stack: List[Any], calling_account: Optional[Account] = None) -> List[NeoInvokeResult]:
        no_iterations = min(len(new_result_stack), len(self._pending_results)) + 1
        pending_results = self._pending_results.copy()

        for index in range(1, no_iterations):
            invoke_result = self._pending_results.pop(0)
            invoke_result.invoke._invoker = calling_account
            invoke_result._result = new_result_stack[-index]

        return pending_results
